/* -*- mode: C; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
/*
 * Copyright (C) 2009-2011  Tiger Soldier <tigersoldier@gmail.com>
 *
 * This file is part of OSD Lyrics.
 * 
 * OSD Lyrics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OSD Lyrics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>. 
 */
/**
 * @file   ol_commands.h
 * @author Tiger Soldier <tigersoldi@gmail.com>
 * @date   Sun Aug 16 16:17:41 2009
 * 
 * @brief  Common operation on OSD Lyrics
 * 
 * 
 */
#ifndef _OL_COMMANDS_H_
#define _OL_COMMANDS_H_

/** 
 * @brief switch lock/unlock status on OSD windows
 * 
 */
void ol_osd_lock_unlock ();

/** 
 * @brief switch show/hide OSD
 * 
 */
void ol_osd_switch_display ();

#endif /* _OL_COMMANDS_H_ */
